# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import unittest
from pydub.exceptions import CouldntDecodeError

from neon_utils.file_utils import *

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
AUDIO_PATH = os.path.join(ROOT_DIR, "tests", "audio_files")
README_PATH = os.path.join(ROOT_DIR, "tests", "readme_files")


class FileUtilTests(unittest.TestCase):

    __audio_data = None

    def test_encode_file(self):
        byte_string = encode_file_to_base64_string(os.path.join(ROOT_DIR, "LICENSE.md"))
        self.assertIsInstance(byte_string, str)

    def test_write_encoded_file(self):
        byte_string = encode_file_to_base64_string(os.path.join(ROOT_DIR, "LICENSE.md"))
        output_path = os.path.join(ROOT_DIR, "tests", "LICENSE.md")
        output_file = decode_base64_string_to_file(byte_string, output_path)
        self.assertEqual(output_path, output_file)
        with open(os.path.join(ROOT_DIR, "LICENSE.md"), "r") as original:
            original_text = original.read()
        with open(output_file, "r") as duplicated:
            duplicate_text = duplicated.read()
        self.assertEqual(original_text, duplicate_text)
        os.remove(output_file)

    def test_get_most_recent_file_in_dir(self):
        newest = get_most_recent_file_in_dir(ROOT_DIR)
        self.assertIsInstance(newest, str)
        print(newest)
        self.assertTrue(os.path.exists(newest))

        newest_py = get_most_recent_file_in_dir(os.path.join(ROOT_DIR, "*.py"))
        self.assertIsInstance(newest_py, str)
        self.assertTrue(os.path.isfile(newest_py))

        newest_dne = get_most_recent_file_in_dir(os.path.join(ROOT_DIR, "*.fake"))
        self.assertIsNone(newest_dne)

    def test_get_most_recent_file_in_dir_with_ext(self):
        newest = get_most_recent_file_in_dir(ROOT_DIR, "py")
        self.assertTrue(os.path.isfile(newest))

        newest_dne = get_most_recent_file_in_dir(ROOT_DIR, "fake")
        self.assertIsNone(newest_dne)

    def test_get_wav_as_wav(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.wav")
        wav_data = get_file_as_wav(test_file, 16000)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 16000)

        wav_data = get_file_as_wav(test_file, 44100)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 44100)

    def test_file_stream_wav(self):
        test_file = os.path.join(AUDIO_PATH, "stop.wav")
        stream = get_audio_file_stream(test_file)
        self.assertEqual(stream.sample_rate, 16000)

        stream = get_audio_file_stream(test_file, 44100)
        self.assertEqual(stream.sample_rate, 44100)

    def test_get_mp3_as_wav(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.mp3")
        wav_data = get_file_as_wav(test_file, 16000)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 16000)

        wav_data = get_file_as_wav(test_file, 44100)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 44100)

    def test_file_stream_mp3(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.mp3")
        stream = get_audio_file_stream(test_file)
        self.assertEqual(stream.sample_rate, 16000)

        stream = get_audio_file_stream(test_file, 44100)
        self.assertEqual(stream.sample_rate, 44100)

    def test_get_flac_as_wav(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.flac")
        wav_data = get_file_as_wav(test_file, 16000)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 16000)

        wav_data = get_file_as_wav(test_file, 44100)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 44100)

    def test_file_stream_flac(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.flac")
        stream = get_audio_file_stream(test_file)
        self.assertEqual(stream.sample_rate, 16000)

        stream = get_audio_file_stream(test_file, 44100)
        self.assertEqual(stream.sample_rate, 44100)

    def test_get_ogg_as_wav(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.ogg")
        wav_data = get_file_as_wav(test_file, 16000)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 16000)

        wav_data = get_file_as_wav(test_file, 44100)
        self.assertIsInstance(wav_data, wave.Wave_read)
        self.assertEqual(wav_data.getframerate(), 44100)

    def test_file_stream_ogg(self):
        test_file = os.path.join(AUDIO_PATH, "testing 1 2 3.ogg")
        stream = get_audio_file_stream(test_file)
        self.assertEqual(stream.sample_rate, 16000)

        stream = get_audio_file_stream(test_file, 44100)
        self.assertEqual(stream.sample_rate, 44100)

    def test_file_stream_invalid_path(self):
        with self.assertRaises(FileNotFoundError):
            get_audio_file_stream(os.path.join(AUDIO_PATH, "nothing"))

    def test_file_stream_invalid_type(self):
        with self.assertRaises(Exception):
            get_audio_file_stream(os.path.join(ROOT_DIR, "README.md"))

    def test_invalid_path_as_wav(self):
        with self.assertRaises(FileNotFoundError):
            get_file_as_wav(os.path.join(AUDIO_PATH, "nothing"), 44100)

    def test_invalid_file_as_wav(self):
        with self.assertRaises(CouldntDecodeError):
            get_file_as_wav(os.path.join(ROOT_DIR, "README.md"), 44100)

    def test_get_file_as_wav_no_ext(self):
        wav_data = get_file_as_wav(os.path.join(AUDIO_PATH, "testing 1 2 3"), 44100)
        self.assertEqual(wav_data.getframerate(), 44100)
        self.assertEqual(wav_data.getsampwidth(), 2)

    def test_get_file_as_wav_change_width(self):
        wav_data = get_file_as_wav(os.path.join(AUDIO_PATH, "testing 1 2 3 (sw4).wav"), 44100)
        self.assertEqual(wav_data.getframerate(), 44100)
        self.assertEqual(wav_data.getsampwidth(), 2)

    def test_resolve_neon_resource_file_valid(self):
        start_listening = resolve_neon_resource_file("snd/start_listening.wav")
        self.assertTrue(os.path.isfile(start_listening))

    def test_resolve_neon_resource_file_invalid(self):
        self.assertIsNone(resolve_neon_resource_file("invalid/file"))

    def test_01_audio_bytes_from_file(self):
        test_path = "testing 1 2 3 (sw4).wav"
        audio_data = audio_bytes_from_file(os.path.join(AUDIO_PATH, test_path))
        self.assertIsInstance(audio_data, dict)
        self.__class__.__audio_data = audio_data
        self.assertTrue(len(audio_data.get('data', [])) > 0)

    def test_02_audio_bytes_to_file(self):
        output_file = "testing output 1 2 3 (sr4).wav"
        self.assertIsNotNone(self.__class__.__audio_data)
        audio_path = audio_bytes_to_file(os.path.join(AUDIO_PATH, output_file),
                                         audio_data=self.__audio_data['data'],
                                         sample_rate=self.__audio_data['sample_rate'])
        self.assertTrue(os.path.exists(audio_path))
        os.remove(audio_path)
        self.assertFalse(os.path.exists(audio_path))

    def test_parse_neon_skill_readme_file_valid(self):
        str_sections = ("title", "icon", "summary", "description", "contact support", "category")
        list_sections = ("examples", "credits", "categories", "tags")
        neon_data = parse_skill_readme_file(os.path.join(README_PATH, "neon_readme.md"))
        self.assertIsInstance(neon_data, dict)
        self.assertEqual(set(neon_data.keys()), {"title", "icon", "summary", "description", "examples",
                                                 "contact support", "credits", "category", "tags", "categories"})

        for val in neon_data.values():
            if val in str_sections:
                self.assertIsInstance(val, str)
            elif val in list_sections:
                self.assertIsInstance(val, list)

        self.assertEqual(neon_data["title"], "Demo")
        self.assertEqual(neon_data["icon"], "https://0000.us/klatchat/app/files/neon_images/icons/neon_skill.png")
        self.assertEqual(neon_data["summary"], "Skill to demo Neon capabilities")
        self.assertEqual(neon_data["description"],
                         "The demo skill will prompt on first run if you'd like to see a demo. User may accept or "
                         "decline and optionally choose to be asked again on next run. The demo may also be run at "
                         "any time via intent.")
        self.assertEqual(neon_data["contact support"],
                         "Use the [link](https://neongecko.com/ContactUs) or [submit an issue on GitHub]"
                         "(https://help.github.com/en/articles/creating-an-issue)")
        self.assertEqual(neon_data["category"], "Information")

        self.assertEqual(neon_data["examples"], ["show me the demo"])
        self.assertEqual(set(neon_data["credits"]), {"NeonGeckoCom", "NeonDaniel"})
        self.assertEqual(neon_data["categories"], ["Information"])
        self.assertEqual(neon_data["tags"], ["NeonGecko", "NeonAI", "Demo"])

        for readme in ("legacy_neon_readme.md", "complex_neon_readme.md"):
            neon_data = parse_skill_readme_file(os.path.join(README_PATH, readme))
            self.assertIsInstance(neon_data, dict)
            for val in neon_data.values():
                if val in str_sections:
                    self.assertIsInstance(val, str)
                elif val in list_sections:
                    self.assertIsInstance(val, list)

    def test_parse_skill_readme_file_invalid(self):
        with self.assertRaises(FileNotFoundError):
            parse_skill_readme_file(os.path.join(README_PATH, "invalid_path"))

        with self.assertRaises(FileNotFoundError):
            parse_skill_readme_file(README_PATH)

        with self.assertRaises(ValueError):
            parse_skill_readme_file("")

    def test_check_path_permissions(self):
        with self.assertRaises(FileNotFoundError):
            check_path_permissions("/test/fnf")
        self.assertEqual(check_path_permissions("/opt"), (True, False, True))
        self.assertEqual(check_path_permissions("/tmp"), (True, True, True))
        self.assertEqual(check_path_permissions("~/"), (True, True, True))

    def test_check_path_is_writable(self):
        self.assertFalse(path_is_read_writable("/test/fnf"))
        self.assertFalse(path_is_read_writable("/opt"))
        self.assertFalse(path_is_read_writable("/opt/dne_test_dir"))
        self.assertTrue(path_is_read_writable("/tmp"))
        self.assertTrue(path_is_read_writable("~/"))
        self.assertTrue(path_is_read_writable("~/.dne_test_dir"))
        self.assertFalse(path_is_read_writable('/'))
        unwritable_path = os.path.join(ROOT_DIR, "tests", "configuration", "unwritable_path")
        self.assertFalse(path_is_read_writable(unwritable_path))

    def test_create_file(self):
        valid_file = os.path.expanduser("~/test_pass")
        with self.assertRaises(PermissionError):
            create_file("/test_fail")
        create_file("~/test_pass")
        self.assertTrue(os.path.isfile(valid_file))
        os.remove(valid_file)

    def test_load_commented_file(self):
        test_dir = os.path.join(os.path.dirname(__file__), "commented_files")
        test_file = os.path.join(test_dir, "test_commented_file.txt")

        contents = load_commented_file(test_file)
        self.assertEqual(contents, "Some valid text\n//Java commented line")

        contents = load_commented_file(test_file, '//')
        self.assertEqual(contents, "# Python Commented Line\nSome valid text")

        contents = load_commented_file(test_file, '/')
        self.assertEqual(contents, "# Python Commented Line\nSome valid text")

        contents = load_commented_file(test_file, '/**')
        self.assertEqual(contents, "# Python Commented Line\nSome valid text\n"
                                   "//Java commented line")


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import re
import importlib.util
from typing import Tuple, Optional, List
from tempfile import mkstemp

import pkg_resources
import sysconfig

from os.path import exists, join, expanduser, isdir
from ovos_utils.log import LOG, deprecated


def parse_version_string(ver: str) -> Tuple[int, int, int, Optional[int]]:
    """
    Parse a semver string into its component versions as ints
    :param ver: Version string to parse
    :returns: Tuple major, minor, patch, Optional(revision)
    """
    parts = ver.split('.')
    major = int(parts[0])
    minor = int(parts[1]) if len(parts) > 1 else 0
    patch = parts[2] if len(parts) > 2 else '0'
    if not patch.isnumeric():
        patch, alpha = re.split(r"\D+", patch, 1)
        alpha = int(alpha)
    else:
        alpha = None
    patch = int(patch)
    return major, minor, patch, alpha


def get_package_version_spec(pkg: str):
    """
    Locate an installed package and return its reported version
    :param pkg: string package name to locate
    :returns: Version string as reported by pkg_resources
    :raises: ModuleNotFoundError if requested package isn't installed
    """
    try:
        return pkg_resources.get_distribution(pkg).version
    except pkg_resources.DistributionNotFound:
        raise ModuleNotFoundError(f"{pkg} not found")


def get_package_dependencies(pkg: str):
    """
    Get the dependencies for an installed package
    :param pkg: string package name to evaluate
    :returns: list of string dependencies (equivalent to requirements.txt)
    :raises ModuleNotFoundError if requested package isn't installed
    """
    try:
        constraints = pkg_resources.working_set.by_key[pkg].requires()
        constraints_spec = [str(c).split('[', 1)[0] for c in constraints]
        LOG.debug(constraints_spec)
        return constraints_spec
    except KeyError:
        raise ModuleNotFoundError(f"{pkg} not found")


@deprecated("Reference `neon_core.version.__version__` directly", "2.0.0")
def get_packaged_core_version() -> str:
    """
    Get the version of the packaged core in use.
    Supports Neon, Mycroft, and OVOS default packages.
    Returns:
        Version of the installed core package
    """
    if importlib.util.find_spec("neon-core"):
        return get_package_version_spec("neon-core")
    elif importlib.util.find_spec("mycroft-core"):
        return get_package_version_spec("mycroft-core")
    elif importlib.util.find_spec("mycroft-lib"):
        return get_package_version_spec("mycroft-lib")
    raise ImportError("No Core Package Found")


@deprecated("Reference `neon_core.version.__version__` directly", "2.0.0")
def get_neon_core_version() -> str:
    """
    Gets the current version of the installed Neon Core.
    Returns:
        Version of the available/active Neon Core or
        0.0 if no release info is found
    """
    try:
        from neon_core.version import __version__
        return __version__
    except ImportError:
        pass
    try:
        return get_packaged_core_version()
    except ImportError:
        pass
    return "0.0"


@deprecated("Use system modules to locate packages and resources", "2.0.0")
def get_core_root():
    """
    Depreciated 2020.09.01
    :return:
    """
    return get_mycroft_core_root()


@deprecated("Use system modules to locate packages and resources", "2.0.0")
def get_neon_core_root():
    """
    Determines the root of the available/active Neon Core.
    Directory returned is the root of the `neon_core` package
    Returns:
        Path to the 'neon_core' directory
    """
    site = sysconfig.get_paths()['platlib']
    if exists(join(site, 'neon_core')):
        return join(site, 'neon_core')
    for p in [path for path in sys.path if path != ""]:
        if exists(join(p, "neon_core")):
            return join(p, "neon_core")
        if re.match(".*/lib/python.*/site-packages", p):
            clean_path = "/".join(p.split("/")[0:-4])
            if exists(join(clean_path, "neon_core")):
                return join(clean_path, "neon_core")
            elif exists(join(p, "neon_core")):
                return join(p, "neon_core")
    raise FileNotFoundError("Could not determine core directory")


@deprecated("Use system modules to locate packages and resources", "2.0.0")
def get_mycroft_core_root():
    """
    Determines the root of the available/active Neon Core.
    Should be the immediate parent directory of 'mycroft' dir
    Returns:
        Path to the core directory containing 'mycroft'
    """
    site = sysconfig.get_paths()['platlib']
    if exists(join(site, 'mycroft')):
        return site
    for p in [path for path in sys.path if path != ""]:
        if exists(join(p, "mycroft")):
            return p
        if re.match(".*/lib/python.*/site-packages", p):
            clean_path = "/".join(p.split("/")[0:-4])
            if exists(join(clean_path, "mycroft")):
                return clean_path
            # TODO: Other packages (Neon Core)? DM
            elif exists(join(p, "mycroft")):
                return p
    raise FileNotFoundError("Could not determine core directory")


def build_skill_spec(skill_dir: str) -> dict:
    """
    Build dict contents of a skill.json file.
    :param skill_dir: path to skill directory to parse
    :returns: dict skill.json spec
    """
    import shutil
    from ovos_skills_manager.local_skill import get_skill_data_from_directory
    from neon_utils.file_utils import parse_skill_readme_file
    from neon_utils.configuration_utils import dict_merge

    def get_skill_license():  # TODO: Implement OSM version of this
        try:
            with open(join(skill_dir, "LICENSE.md")) as f:
                contents = f.read()
        except FileNotFoundError:
            return "Unknown"
        except Exception as e:
            LOG.error(e)
            return "Unknown"
        if "BSD-3" in contents:
            return "BSD-3-Clause"
        if "Apache License" in contents:
            return "Apache 2.0"
        if "Neon AI Non-commercial Friendly License 2.0" in contents:
            return "Neon 2.0"
        if "Neon AI Non-commercial Friendly License" in contents:
            return "Neon 1.0"

    _invalid_skill_data_keys = ("appstore", "appstore_url", "credits",
                                "skill_id")
    _invalid_readme_keys = ("contact support", "details")
    default_skill = {"title": "",
                     "url": "",
                     "summary": "",
                     "short_description": "",
                     "description": "",
                     "examples": [],
                     "desktopFile": False,
                     "warning": "",
                     "systemDeps": False,
                     "requirements": {
                         "python": [],
                         "system": {},
                         "skill": []
                     },
                     "incompatible_skills": [],
                     "platforms": ["i386",
                                   "x86_64",
                                   "ia64",
                                   "arm64",
                                   "arm"],
                     "branch": "master",
                     "license": "",
                     "icon": "",
                     "category": "",
                     "categories": [],
                     "tags": [],
                     "credits": [],
                     "skillname": "",
                     "authorname": "",
                     "foldername": None}

    skill_dir = expanduser(skill_dir)
    if not isdir(skill_dir):
        raise FileNotFoundError(f"Not a Directory: {skill_dir}")
    LOG.debug(f"skill_dir={skill_dir}")
    skill_json = join(skill_dir, "skill.json")
    backup = join(skill_dir, "skill_json.bak")
    shutil.move(skill_json, backup)
    skill_data = get_skill_data_from_directory(skill_dir)
    shutil.move(backup, skill_json)
    skill_data['foldername'] = None
    for key in _invalid_skill_data_keys:
        if key in skill_data:
            skill_data.pop(key)
    readme_data = parse_skill_readme_file(join(skill_dir, "README.md"))
    for key in _invalid_readme_keys:
        if key in readme_data:
            readme_data.pop(key)
    readme_data["short_description"] = readme_data.get("summary")
    readme_data["license"] = get_skill_license()
    readme_data["branch"] = "master"
    skill_data = dict_merge(default_skill, skill_data)
    skill_data["requirements"]["python"].sort()
    return dict(dict_merge(skill_data, readme_data))

def install_packages_from_pip(core_module: str, packages: List[str]) -> int:
    """
    Install a Python package using pip
    :param core_module: string neon core module to install dependency for
    :param packages: List(string) list of packages to install
    :returns: int pip exit code
    """
    import pip
    _, tmp_constraints_file = mkstemp()
    _, tmp_requirements_file = mkstemp()

    install_str = " ".join(packages)

    with open(tmp_constraints_file, 'w', encoding="utf8") as f:
        constraints = '\n'.join(get_package_dependencies(core_module))
        f.write(constraints)
        LOG.info(f"Constraints={constraints}")

    with open(tmp_requirements_file, "w", encoding="utf8") as f:
        for pkg in packages:
            f.write(f"{pkg}\n")

    LOG.info(f"Requested installation of plugins: {install_str}")
    pip_args = ['install', '-r', tmp_requirements_file, '-c', tmp_constraints_file]
    result = pip.main(pip_args) if hasattr(pip, 'main') else pip._internal.main(pip_args)
    
    if result != 0:
        return result
    return 0



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from socket import gethostname
from time import time, strftime
from ovos_bus_client import Message, MessageBusClient
from ovos_utils.log import LOG

from neon_utils.message_utils import dig_for_message


class Stopwatch:
    """
    Provides a stopwatch object compatible with mycroft.metrics Stopwatch.
    """

    def __init__(self, metric_name=None, allow_reporting=False, bus=None):
        """
        Create a stopwatch object with an optional metric_name
        @param metric_name: name of the metric this stopwatch is measuring
        @param allow_reporting: boolean flag to allow this stopwatch to report measured metrics
        @param bus: MessageBusClient object to report metrics with
        """
        self._metric = metric_name
        self._report = allow_reporting
        self._context = dict()
        self._bus = bus
        self.start_time = None
        self.time = None

    def __enter__(self):
        self.start()

    def __exit__(self, typ, val, traceback):
        self.stop()
        try:
            self.report()
        except Exception as e:
            LOG.exception(e)

    def start(self):
        self.start_time = time()

    def stop(self):
        try:
            self.time = time() - self.start_time
        except TypeError:
            LOG.error("stop called before start!")
            self.time = None
        return self.time

    def report(self):
        if self._metric and self._report:
            if not self._bus:
                self._bus = MessageBusClient()
                self._bus.run_in_thread()
            message = dig_for_message() or Message("")
            message.context['timestamp'] = time()
            self._bus.emit(message.forward("neon.metric",
                                           {"name": self._metric,
                                            "duration": self.time}))


def report_metric(name: str, **kwargs):
    """
    Report a metric over the MQ bus.
    :param name: Name of the metric to report
    :param kwargs: Arbitrary data to include with metric report
    """
    # TODO: Deprecate and move to PHAL plugin
    try:
        from neon_utils.mq_utils import send_mq_request
        send_mq_request("/neon_metrics", {**{"name": name}, **kwargs},
                        "neon_metrics_input", expect_response=False)
        return True
    except Exception as e:
        LOG.error(e)
        return False


def announce_connection():
    try:
        from neon_utils.mq_utils import send_mq_request
        from neon_core.version import __version__
        from ovos_config.config import Configuration
        data = {"time": strftime('%Y-%m-%d %H:%M:%S'),
                "name": dict(Configuration()).get("device_name") or "unknown",
                "host": gethostname(),
                "ver": __version__}
        send_mq_request("/neon_metrics", data, "neon_connections_input",
                        expect_response=False)
        return True
    except Exception as e:
        LOG.error(e)
        return False



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from os.path import dirname

from ovos_workshop.decorators.layers import IntentLayers
from ovos_workshop.skills.common_query_skill import CQSMatchLevel, CQSVisualMatchLevel
from ovos_workshop.skills.common_query_skill import CommonQuerySkill as _CQS
from ovos_workshop.decorators.compat import backwards_compat
from ovos_utils.file_utils import resolve_resource_file
from ovos_utils.log import log_deprecation, LOG
from neon_utils.skills.neon_skill import NeonSkill


def is_CQSVisualMatchLevel(match_level):
    log_deprecation("This method is deprecated", "2.0.0")
    return isinstance(match_level, type(CQSVisualMatchLevel.EXACT))


VISUAL_DEVICES = ['mycroft_mark_2']


def handles_visuals(platform):
    log_deprecation("This method is deprecated", "2.0.0")
    return platform in VISUAL_DEVICES


class CommonQuerySkill(NeonSkill, _CQS):
    """Question answering skills should be based on this class.

    The skill author needs to implement `CQS_match_query_phrase` returning an
    answer and can optionally implement `CQS_action` to perform additional
    actions if the skill's answer is selected.

    This class works in conjunction with skill-query which collects
    answers from several skills presenting the best one available.
    """
    def __init__(self, *args, **kwargs):
        log_deprecation("This base class is deprecated. Implement "
                        "`ovos_workshop.skills.common_query_skill."
                        "CommonQuerySkill`", "2.0.0")
        # these should probably be configurable
        self.level_confidence = {
            CQSMatchLevel.EXACT: 0.9,
            CQSMatchLevel.CATEGORY: 0.6,
            CQSMatchLevel.GENERAL: 0.5
        }

        # Manual init of OVOSSkill
        self.private_settings = None
        self._threads = []
        self._original_converse = self.converse
        self.intent_layers = IntentLayers()
        self.audio_service = None

        NeonSkill.__init__(self, *args, **kwargs)

        noise_words_filepath = f"text/{self.lang}/noise_words.list"
        default_res = f"{dirname(dirname(__file__))}/res/text/{self.lang}" \
                      f"/noise_words.list"
        noise_words_filename = \
            resolve_resource_file(noise_words_filepath,
                                  config=self.config_core) or \
            resolve_resource_file(default_res, config=self.config_core)

        self._translated_noise_words = {}
        if noise_words_filename:
            with open(noise_words_filename) as f:
                translated_noise_words = f.read().strip()
            self._translated_noise_words[self.lang] = \
                translated_noise_words.split()

    def __calc_confidence(self, match, phrase, level):
        # Assume the more of the words that get consumed, the better the match
        consumed_pct = len(match.split()) / len(phrase.split())
        if consumed_pct > 1.0:
            consumed_pct = 1.0

        # Add bonus if match has visuals and the device supports them.
        # platform = self.config_core.get('enclosure', {}).get('platform')
        # if is_CQSVisualMatchLevel(level) and handles_visuals(platform):
        #     bonus = 0.1
        # else:
        #     bonus = 0
        bonus = 0

        if int(level) == int(CQSMatchLevel.EXACT):
            return 0.9 + (consumed_pct / 10) + bonus
        elif int(level) == int(CQSMatchLevel.CATEGORY):
            return 0.6 + (consumed_pct / 10) + bonus
        elif int(level) == int(CQSMatchLevel.GENERAL):
            return 0.5 + (consumed_pct / 10) + bonus
        else:
            return 0.0  # should never happen

    def __handle_query_classic(self, message):
        """Message handler for question:action.

        Extracts phrase and data from message forward this to the skills
        CQS_action method.
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        LOG.debug(f"handling for ovos-core 0.0.7")
        phrase = message.data["phrase"]
        data = message.data.get("callback_data")
        # Invoke derived class to provide playback data
        self.CQS_action(phrase, data)
        self.bus.emit(message.forward("mycroft.skill.handler.complete",
                                      {"handler": "common_query"}))

    @backwards_compat(classic_core=__handle_query_classic,
                      pre_008=__handle_query_classic)
    def __handle_query_action(self, message):
        """
        If this skill's response was spoken to the user, this method is called.
        Phrase and callback data from `CQS_match_query_phrase` will be passed
        to the `CQS_action` method.
        @param message: `question:action` message
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        LOG.debug(f"handling for ovos-core 0.0.8")
        phrase = message.data["phrase"]
        data = message.data.get("callback_data") or {}
        if data.get("answer"):
            self.speak(data["answer"])
        else:
            LOG.error(f"no answer provided in: {message.data.keys()}")
        # Invoke derived class to provide playback data
        self.CQS_action(phrase, data)
        self.bus.emit(message.forward("mycroft.skill.handler.complete",
                                      {"handler": "common_query"}))

    def __handle_question_query(self, message):
        # Override ovos-workshop implementation that doesn't pass `message`
        search_phrase = message.data["phrase"]

        # First, notify the requestor that we are attempting to handle
        # (this extends a timeout while this skill looks for a match)
        self.bus.emit(message.response({"phrase": search_phrase,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Now invoke the CQS handler to let the skill perform its search
        result = self.CQS_match_query_phrase(search_phrase, message)

        if result:
            match = result[0]
            level = result[1]
            answer = result[2]
            callback = result[3] if len(result) > 3 else None
            callback["answer"] = answer
            confidence = self.__calc_confidence(match, search_phrase, level)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "answer": answer,
                                            "callback_data": callback,
                                            "conf": confidence}))
        else:
            # Signal we are done (can't handle it)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    @abstractmethod
    def CQS_match_query_phrase(self, phrase, message):
        """Analyze phrase to see if it is a play-able phrase with this skill.

        Needs to be implemented by the skill.

        Arguments:
            phrase (str): User phrase, "What is an aardwark"

        Returns:
            (match, CQSMatchLevel[, callback_data]) or None: Tuple containing
                 a string with the appropriate matching phrase, the PlayMatch
                 type, and optionally data to return in the callback if the
                 match is selected.
        """
        # Derived classes must implement this, e.g.
        return None

    def CQS_action(self, phrase, data):
        """Take additional action IF the skill is selected.

        The speech is handled by the common query but if the chosen skill
        wants to display media, set a context or prepare for sending
        information info over e-mail this can be implemented here.

        Args:
            phrase (str): User phrase uttered after "Play", e.g. "some music"
            data (dict): Callback data specified in match_query_phrase()
        """
        # Derived classes may implement this if they use additional media
        # or wish to set context after being called.
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import uuid

from ovos_utils.log import deprecated, log_deprecation

log_deprecation("This module has moved to neon_mq_connector.utils.client_utils",
                "2.0.0")
try:
    from threading import Event
    from pika.channel import Channel
    from pika.exceptions import ProbableAccessDeniedError, StreamLostError
    from neon_mq_connector.connector import MQConnector
    from ovos_config.config import Configuration
except ImportError:
    raise ImportError("MQ dependencies not available,"
                      " pip install neon-utils[network]")

from neon_utils.logger import LOG
from neon_utils.socket_utils import b64_to_dict

logging.getLogger("pika").setLevel(logging.CRITICAL)

_default_mq_config = {
    "server": "mq.neonaiservices.com",
    "port": 5672,
    "users": {
        "mq_handler": {
            "user": 'neon_api_utils',
            "password": 'Klatchat2021'
        }
    }
}


class NeonMQHandler(MQConnector):
    def __init__(self, config: dict, service_name: str, vhost: str):
        log_deprecation("Import from neon_mq_connector.utils.client_utils",
                        "2.0.0")
        super().__init__(config, service_name)
        self.vhost = vhost
        import pika
        self.connection = pika.BlockingConnection(
            parameters=self.get_connection_params(vhost))


@deprecated("Use `neon_mq_connector.client.send_mq_request`", "2.0.0")
def get_mq_response(vhost: str, request_data: dict, target_queue: str,
                    response_queue: str = None, timeout: int = 30) -> dict:
    return send_mq_request(vhost, request_data, target_queue, response_queue,
                           timeout, True)


@deprecated("Import from neon_mq_connector.utils.client_utils", "2.0.0")
def send_mq_request(vhost: str, request_data: dict, target_queue: str,
                    response_queue: str = None, timeout: int = 30,
                    expect_response: bool = True) -> dict:
    """
    Sends a request to the MQ server and returns the response.
    :param vhost: vhost to target
    :param request_data: data to post to target_queue
    :param target_queue: queue to post request to
    :param response_queue: optional queue to monitor for a response. Generally should be blank
    :param timeout: time in seconds to wait for a response before timing out
    :param expect_response: boolean indicating whether or not a response is expected
    :return: response to request
    """
    response_queue = response_queue or uuid.uuid4().hex

    response_event = Event()
    message_id = None
    response_data = dict()
    config = dict()

    def on_error(thread, error):
        """
        Override default error handler to suppress certain logged errors.
        """
        if isinstance(error, StreamLostError):
            return
        LOG.error(f"{thread} raised {error}")

    def handle_mq_response(channel: Channel, method, _, body):
        """
            Method that handles Neon API output.
            In case received output message with the desired id, event stops
        """
        api_output = b64_to_dict(body)
        api_output_msg_id = api_output.get('message_id', None)
        if api_output_msg_id == message_id:
            LOG.debug(f'MQ output: {api_output}')
            channel.basic_ack(delivery_tag=method.delivery_tag)
            channel.close()
            response_data.update(api_output)
            response_event.set()
        else:
            channel.basic_nack(delivery_tag=method.delivery_tag)
            LOG.debug(f"Ignoring {api_output_msg_id} waiting for {message_id}")

    try:
        # LOG.debug('Creating Neon MQ Handler Instance...')
        config = dict(Configuration()).get('MQ') or _default_mq_config
        # LOG.info(f"MQ Config={config}")
        neon_api_mq_handler = NeonMQHandler(config=config, service_name='mq_handler', vhost=vhost)
        # LOG.debug(f'Established MQ connection: {neon_api_mq_handler.connection}')
        if not neon_api_mq_handler.connection.is_open:
            raise ConnectionError("MQ Connection not established.")

        if expect_response:
            neon_api_mq_handler.register_consumer('neon_output_handler',
                                                  neon_api_mq_handler.vhost, response_queue,
                                                  handle_mq_response, on_error, auto_ack=False)
            neon_api_mq_handler.run_consumers()
            request_data['routing_key'] = response_queue

        message_id = neon_api_mq_handler.emit_mq_message(connection=neon_api_mq_handler.connection,
                                                         queue=target_queue,
                                                         request_data=request_data,
                                                         exchange='')
        LOG.debug(f'Sent request with keys: {request_data.keys()}')

        if expect_response:
            response_event.wait(timeout)
            if not response_event.is_set():
                LOG.error(f"Timeout waiting for response to: {message_id} on {response_queue}")
            neon_api_mq_handler.stop_consumers()
    except ProbableAccessDeniedError:
        raise ValueError(f"{vhost} is not a valid endpoint for {config.get('users').get('mq_handler').get('user')}")
    except Exception as ex:
        LOG.error(f'Exception occurred while resolving Neon API: {ex}')
    return response_data



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import base64
import inspect

from enum import Enum
from time import time
from typing import Optional, Union
from ovos_bus_client import Message
from ovos_utils.log import deprecated

from neon_utils.logger import LOG


class MessageKind(Enum):
    SPEAK = "speak"
    DATA = "skills:execute.utterance"
    EXECUTE = "skills:execute.utterance"


class EncodingError(ValueError):
    """Exception to indicate an invalid Encoding"""


def request_from_mobile(message: Message) -> bool:
    """
    Check if a request is from a mobile device
    Args:
        message: Message object associated with request

    Returns:
        True if message is from a mobile app, else False
    """
    return message.context.get("mobile", False)


def get_message_user(message: Message) -> Optional[str]:
    """
    Get the user associated with a message
    Args:
        message: Message object associated with request

    Returns:
        Username associated with message
    """
    if not message:
        raise ValueError
    if not hasattr(message, "context"):
        raise AttributeError(type(message))
    return message.context.get("username")


def encode_bytes_to_b64_string(data: bytes, charset: str = "utf-8") -> str:
    """
    Encode a bytes object into a string that can be passed on the messagebus
    :param data: bytes data to be encoded
    :param charset: character set to decode b64 bytes (https://docs.python.org/3/library/codecs.html#standard-encodings)
    :return: base64 encoded string
    """
    if not data or not isinstance(data, bytes):
        raise ValueError(f"Invalid data provided to be encoded. type={type(data)}")

    encoded = base64.b64encode(data)

    try:
        bytestr = encoded.decode(charset)
    except LookupError:
        raise EncodingError(f"Invalid charset provided: {charset}")

    return bytestr


def decode_b64_string_to_bytes(data: str, charset: str = "utf-8") -> bytes:
    """
    Decodes a base64-encoded string to bytes
    :param data: string encoded data to decode
    :param charset: character set of b64 string (https://docs.python.org/3/library/codecs.html#standard-encodings)
    :return: decoded bytes object
    """
    if not data or not isinstance(data, str):
        raise ValueError(f"Invalid data provided to be encoded. type={type(data)}")

    try:
        byte_data = data.encode(charset)
    except LookupError:
        raise EncodingError(f"Invalid charset provided: {charset}")

    encoded = base64.b64decode(byte_data)

    if not encoded:
        raise EncodingError(f"Invalid charset provided for data. charset={charset}")

    return encoded


def dig_for_message(max_records: int = 10) -> Optional[Message]:
    """
    Dig Through the stack for message. Looks at the current stack for a passed argument of type 'Message'
    :param max_records: Maximum number of stack records to look through
    :return: Message if found in args, else None
    """
    stack = inspect.stack()[1:]  # First frame will be this function call
    stack = stack if len(stack) <= max_records else stack[:max_records]
    for record in stack:
        args = inspect.getargvalues(record.frame)
        if args.args:
            for arg in args.args:
                if isinstance(args.locals[arg], Message):
                    return args.locals[arg]
    return None


def resolve_message(function):
    """
    Decorator to try and fill an optional `message` kwarg
    """
    def wrapper(*args, **kwargs):
        params = inspect.signature(function).parameters
        if not any([param in params for param in ("message", "kwargs")]):
            LOG.warning("Decorated function does not expect a `message`")
            return function(*args, **kwargs)

        if "message" in params.keys() and len(args):
            i = 0
            for param in params:
                args = list(args)
                # Check if 'message' is filled by an arg
                if param == "message" and i < len(args):
                    if not args[i]:
                        args[i] = dig_for_message(50)
                        return function(*args, **kwargs)
                i += 1

        if not kwargs.get("message") and not any([arg for arg in args if
                                                  isinstance(arg, Message)]):
            if LOG.diagnostic_mode:
                call = inspect.stack()[1]
                module = inspect.getmodule(call.frame)
                name = module.__name__ if module else call.filename
                LOG.debug(f"Digging for requested message arg - "
                          f"{name}:{call.lineno}")
            message = dig_for_message(50)
            kwargs["message"] = message
        return function(*args, **kwargs)

    return wrapper


@resolve_message
def request_for_neon(message: Message = None,
                     activation_voc: Optional[str] = "neon",
                     voc_match: Optional[callable] = None,
                     ww_enabled: bool = True) -> bool:
    """
    Check if a wake word or activation word is present in a request
    :param message: Message to evaluate
    :param activation_voc: vocab of activation word for the assistant
    :param voc_match: `skill.voc_match` method to check for `activation_voc`
    :param ww_enabled: True if WW is required for STT system-wide
    :returns: True if the message contains an activation word
    """
    if not message:
        raise ValueError("message not provided and not resolvable")
    if message.context.get('neon_should_respond'):
        return True
    if activation_voc in message.data:
        return True

    server_request = message.context.get('klat_data')

    if ww_enabled and not server_request:
        # TODO: WW should add to message context instead DM
        return True

    # Check Klat conversation context
    if server_request and \
            server_request.get('title', '').startswith("!PRIVATE"):
        return True

    # Check for voc_match if method provided
    try:
        if voc_match and voc_match(message.data.get('utterance', '').lower(),
                                   activation_voc):
            return True
    except FileNotFoundError as e:
        LOG.error(e)

    return False


@resolve_message
def build_message(kind: Union[MessageKind, str], utt: str, message: Message,
                  speaker: Optional[dict] = None) -> Message:
    """
    Build a message object for skill execution or TTS output
    :param kind: "neon speak" or "execute"
    :param utt: string to emit
    :param message: incoming message object
    :param speaker: speaker data dictionary
    :return: Message object
    """
    from neon_utils.user_utils import get_user_prefs
    if isinstance(kind, str):
        kind = MessageKind.EXECUTE if kind == "execute" else \
            MessageKind.DATA if kind == "skill_data" else MessageKind.SPEAK

    default_speech = get_user_prefs(message)["speech"]
    speaker = speaker or {"name": "Neon",
                          "language": default_speech["tts_language"],
                          "gender": default_speech["tts_gender"],
                          "voice": default_speech["neon_voice"],
                          "override_user": True}
    if speaker.get("language"):
        speaker['override_user'] = True

    if kind in (MessageKind.EXECUTE, MessageKind.DATA):
        return message.reply(kind.value, {
            "utterances": [utt.lower()],
            "lang": message.data.get("lang", "en-US"),
            "session": None,
            "ident": time(),
            "speaker": speaker
        }, {
            "neon_should_respond": True,
            "cc_data": {"request": utt,
                        "emit_response": kind == MessageKind.DATA,
                        "execute_from_script": True
                        }
        })
    elif kind == MessageKind.SPEAK:
        added_context = {"cc_data": message.context.get("cc_data", {})}
        added_context["cc_data"]["request"] = utt

        return message.reply(kind.value, {"utterance": utt,
                                          "lang": message.data.get("lang",
                                                                   "en-US"),
                                          "speaker": speaker
                                          }, {**message.context,
                                              **added_context})


@deprecated("Neon will always provide a response to API inputs.", "2.0.0")
@resolve_message
def neon_must_respond(message: Message = None) -> bool:
    """
    Checks if Neon must respond to an utterance (i.e. a server request)
    :param message: message associated with user request
    :returns: True if Neon must provide a response to this request
    """
    if not message:
        return False
    if "klat_data" in message.context:
        title = message.context.get("klat_data", {}).get("title", "")
        LOG.debug(message.data.get("utterance"))
        if message.data.get("utterance", "").startswith(
                "Welcome to your private conversation"):
            return False
        if title.startswith("!PRIVATE:"):
            if ',' in title:
                users = title.split(':')[1].split(',')
                for idx, val in enumerate(users):
                    users[idx] = val.strip()
                if len(users) == 2 and "Neon" in users:
                    # Private with Neon
                    # LOG.debug("DM: Private Conversation with Neon")
                    return True
                elif message.data.get("utterance",
                                      "").lower().startswith("neon"):
                    # Message starts with "neon", must respond
                    return True
            else:
                # Solo Private
                return True
    if message.context.get("client") in ("mana", "mq_api"):
        # API requests should always return a response
        return True
    return False



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from os.path import join
from ovos_utils.log import LOG, log_deprecation
from neon_utils.file_utils import resolve_neon_resource_file
from ovos_workshop.skills.base import SkillGUI as _SkillGUI


class SkillGUI(_SkillGUI):
    def __init__(self, skill):
        log_deprecation("Implement ovos_utils.gui.GUIInterface directly",
                        "2.0.0")
        super().__init__(skill)
        self.serving_http = skill.config_core["skills"].get(
            "run_gui_file_server", False)

    def _pages2uri(self, page_names):
        # Convert pages to full reference
        page_urls = []
        for name in page_names:
            if name.startswith("SYSTEM"):
                if self.serving_http:
                    page = f"{self.remote_url}/system/ui/{name}"
                else:
                    page = resolve_neon_resource_file(join('ui', name))
            else:
                page = self.skill.find_resource(name, 'ui')
                if self.serving_http:
                    parts = page.split('/')
                    LOG.debug(parts)
                    path_suffix = join(*parts[(parts.index('ui') - 1):])
                    LOG.debug(path_suffix)
                    page = join(self.remote_url, "skills", path_suffix)
                    LOG.debug(page)
            if page:
                if self.remote_url and not self.serving_http:
                    page_urls.append(self.remote_url + "/" + page)
                elif "://" in page:
                    page_urls.append(page)
                else:
                    page_urls.append("file://" + page)
            else:
                # Unresolved files, try base implementation
                return super()._pages2uri(page_names)
        return page_urls



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from lingua_franca import parse, set_default_lang

from ovos_workshop.skills.mycroft_skill import MycroftSkill


def numeric_confirmation_validator(confirmation_num: str, lang: str = "en"):
    """
    Returns a validator method that returns true if the confirmation_num is in the passed string.
     The confirmation_num must be the only number in the utterance (something 123 != 1 something 123).
    :param confirmation_num: number to locate in utterance
    :param lang: language to use for extracting number from strings
    :return: True if the confirmation is in the spoken utterance.
    """
    if not isinstance(confirmation_num, str):
        raise TypeError(f"Expected str, got {type(confirmation_num)}")
    if not confirmation_num.isnumeric():
        raise ValueError(f"{confirmation_num} is not numeric")

    set_default_lang(lang)

    def wrapped_validator(utt):
        spoken_num = "".join([str(round(n)) for n in parse.extract_numbers(utt)])
        return confirmation_num == spoken_num
    return wrapped_validator


def string_confirmation_validator(confirmation_str: str):
    """
    Returns a validator method that returns true if the confirmation_str is in the passed string.
    :param confirmation_str: substring to locate in utterance
    :return: True if the confirmation is in the spoken utterance.
    """
    if not isinstance(confirmation_str, str):
        raise TypeError(f"Expected str, got {type(confirmation_str)}")
    if not confirmation_str:
        raise ValueError(f"Got empty confirmation string")

    def wrapped_validator(utt):
        return confirmation_str in utt
    return wrapped_validator


def voc_confirmation_validator(confirmation_voc: str, skill: MycroftSkill):
    """
    Returns a validator method that returns true if the confirmation_voc vocab resource is in the passed string.
    :param confirmation_voc: vocab resource to locate in utterance
    :param skill: skill with access to the passed vocab
    :return: True if the confirmation is in the spoken utterance.
    """
    if not isinstance(confirmation_voc, str):
        raise TypeError(f"Expected str, got {type(confirmation_voc)}")
    if not skill.find_resource(f"{confirmation_voc}.voc"):
        raise FileNotFoundError(f"Could not locate requested vocab: {confirmation_voc}")

    def wrapped_validator(utt):
        return skill.voc_match(utt, confirmation_voc)
    return wrapped_validator




# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
import socket

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

TCP_TEST = "mq.neonaiservices.com"


class NetUtilTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        from socket import socket
        cls.base_socket = socket

    @classmethod
    def tearDownClass(cls) -> None:
        socket.socket = cls.base_socket

    def setUp(self):
        socket.socket = self.base_socket

    def test_get_ip_address_valid(self):
        from neon_utils.net_utils import get_ip_address

        ip_addr = get_ip_address()
        self.assertIsInstance(ip_addr, str)

    def test_get_ip_address_offline(self):
        def mock_socket(*args, **kwargs):
            raise OSError("Socket is disabled!")
        socket.socket = mock_socket
        from neon_utils.net_utils import get_ip_address

        with self.assertRaises(OSError):
            get_ip_address()

    def test_get_adapter_info(self):
        from neon_utils.net_utils import get_adapter_info
        try:
            info = get_adapter_info()
            self.assertIsInstance(info, dict)
            self.assertIsInstance(info.get("mac"), str)
            self.assertEqual(len(info["mac"]), 17)
            self.assertIsInstance(info.get("ipv4"), str)
            self.assertEqual(len(info["ipv4"].split('.')), 4)
            self.assertIsInstance(info.get("ipv6"), str)
            print(info["ipv6"])
            self.assertGreater(info["ipv6"].count(':'), 3)
        except IndexError:
            print("No Connection")

    def test_get_adapter_fail(self):
        from neon_utils.net_utils import get_adapter_info

        with self.assertRaises(IndexError):
            get_adapter_info("FAIL")

    def test_check_url_connection_valid_online(self):
        from neon_utils.net_utils import check_url_response

        self.assertTrue(check_url_response())
        self.assertTrue(check_url_response("google.com"))
        self.assertTrue(check_url_response("https://github.com"))

    def test_check_url_connection_invalid_schema(self):
        from neon_utils.net_utils import check_url_response

        with self.assertRaises(ValueError):
            check_url_response("smb://google.com")
        with self.assertRaises(ValueError):
            check_url_response("ssh://github.com")

    def test_check_url_connection_invalid_args(self):
        from neon_utils.net_utils import check_url_response

        with self.assertRaises(ValueError):
            check_url_response("")
        with self.assertRaises(ValueError):
            check_url_response(123)
        with self.assertRaises(ValueError):
            check_url_response(None)

    def test_check_url_connection_valid_offline(self):
        def mock_socket(*args, **kwargs):
            raise ConnectionError("Socket is disabled!")
        socket.socket = mock_socket
        from neon_utils.net_utils import check_url_response

        self.assertFalse(check_url_response())

    def test_check_url_connection_invalid_url(self):
        from neon_utils.net_utils import check_url_response

        self.assertFalse(check_url_response(f"https://{TCP_TEST}"))

    def test_check_online_valid_online(self):
        from neon_utils.net_utils import check_online
        self.assertTrue(check_online())
        self.assertTrue(check_online(("google.com", "github.com")))
        self.assertTrue(check_online((TCP_TEST, "google.com")))
        self.assertTrue(check_online(("", "google.com")))

    def test_check_online_invalid_offline(self):
        from neon_utils.net_utils import check_online
        self.assertFalse(check_online((TCP_TEST,)))
        self.assertFalse(check_online(("",)))

    def test_check_online_valid_offline(self):
        def mock_socket(*args, **kwargs):
            raise ConnectionError("Socket is disabled!")
        socket.socket = mock_socket
        from neon_utils.net_utils import check_online
        self.assertFalse(check_online())
        self.assertFalse(check_online(("google.com", "github.com")))

    def test_check_online_invalid_params(self):
        from neon_utils.net_utils import check_online
        with self.assertRaises(ValueError):
            check_online(None)
        with self.assertRaises(ValueError):
            check_online("google.com")
        with self.assertRaises(ValueError):
            check_online(123)

    def test_check_port_is_open(self):
        from neon_utils.net_utils import check_port_is_open
        self.assertTrue(check_port_is_open(TCP_TEST, 5672))
        self.assertFalse(check_port_is_open("www.neon.ai", 5672))


if __name__ == '__main__':
    unittest.main()
